/*****************************************************************
**
**	@(#) zkt-soaserial.c  (c) Oct 2015  Holger Zuleger  hznet.de
**
**	A small utility to print out the (unixtime) soa serial
**	number in a human readable form
**	Option -u added to update (increment) the serial number
**
**	This is a modified version of the zkt-soaserial command from ZKT 1.2.3
**
**	Copyright (c) Oct 2007, Holger Zuleger HZnet. All rights reserved.
**
**	This software is open source.
**
**	Redistribution and use in source and binary forms, with or without
**	modification, are permitted provided that the following conditions
**	are met:
**
**	Redistributions of source code must retain the above copyright notice,
**	this list of conditions and the following disclaimer.
**
**	Redistributions in binary form must reproduce the above copyright notice,
**	this list of conditions and the following disclaimer in the documentation
**	and/or other materials provided with the distribution.
**
**	Neither the name of Holger Zuleger HZnet nor the names of its contributors may
**	be used to endorse or promote products derived from this software without
**	specific prior written permission.
**
**	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
**	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
**	TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
**	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
**	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
**	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
**	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
**	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
**	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
**	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
**	POSSIBILITY OF SUCH DAMAGE.
**
*****************************************************************/
# include <stdio.h>
# include <string.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <time.h>
# include <utime.h>
# include <assert.h>
# include <stdlib.h>
# include <ctype.h>

# include "debug.h"
# include "misc.h"


const char *progname;
int	verbose;

static	char	guess_serialfmt (unsigned long serial);
static	int	inc_serial (const char *fname);
static	int	inc_soa_serial (FILE *fp);
static	int	is_soa_rr (const char *line);
static	const	char	*strfindstr (const char *str, const char *search);

static	int	read_serial_fromfile (const char *fname, unsigned long *serial);
static	void	printserial (const char *fname, unsigned long serial);
static	void	usage (const char *msg);

/****************************************************************
**
**	int	inc_serial (filename)
**
**	This function depends on a special syntax formating the
**	SOA record in the zone file!!
**
**	To match the SOA record, the SOA RR must be formatted
**	like this:
**	@ [ttl]   IN  SOA <master.fq.dn.> <hostmaster.fq.dn.> (
**	<SPACEes or TABs>      1234567890; serial number 
**	<SPACEes or TABs>      86400	 ; other values
**				...
**	The space from the first digit of the serial number to
**	the first none white space char or to the end of the line
**	must be at least 10 characters!
**	So you have to left justify the serial number in a field
**	of at least 10 characters like this:
**	<SPACEes or TABs>      1         ; Serial 
**
**	Since ZKT 1.1.0 single line SOA records are also supported
**
****************************************************************/
static	int	inc_serial (const char *fname)
{
	FILE	*fp;
	char	buf[4095+1];
	int	error;
	int	serial_pos;

	if ( (fp = fopen (fname, "r+")) == NULL )
		return -1;

		/* read until the line matches the beginning of a soa record ... */
	while ( fgets (buf, sizeof buf, fp) )
	{
		dbg_val ("inc_serial() checking line for SOA RR \"%s\"\n", buf);
		serial_pos = is_soa_rr (buf);
		if ( serial_pos )	/* SOA record found ? */
			break;
	}

	if ( feof (fp) )
	{
		fclose (fp);
		return -2;
	}
	dbg_val ("serial_pos = %d\n", serial_pos);
	if (serial_pos > 1 )	/* if we found a single line SOA RR */
		fseek (fp, -(long)serial_pos, SEEK_CUR);	/* go back to the beginning of the line */

	error = inc_soa_serial (fp);	/* .. inc soa serial no ... */
	dbg_val ("inc_soa_serial() returns %d\n", error);

	if ( fclose (fp) != 0 )		/* close the zone file in any case */
		return -5;
	return error;
}

/*****************************************************************
**
**	check if line is the beginning of a SOA RR record, thus
**	containing the string "IN .* SOA" and ends with a '('
**	(multiline record) or is a single line record.
**
**	returns 1 if it is a multi line record (for compability to
**	the old function) or the position of the serial number
**	field counted from the end of the line 
**
*****************************************************************/
static	int	is_soa_rr (const char *line)
{
	const	char	*p;
	const	char	*soa_p;

	assert ( line != NULL );

			/* line contains "IN" and "SOA" ? */
	if ( (p = strfindstr (line, "IN")) && (soa_p = strfindstr (p+2, "SOA")) )
	{
		int	len = strlen (line);

		/* check for multiline record */
		p = line + len - 1;
		while ( p > line && isspace (*p) )
			p--;
		if ( *p == '(' )	/* last character must be a '(' to start a multi line record */
			return 1;

		/* line is single line record */
		p = soa_p + 3;			/* start just behind the SOA string */
		dbg_val1 ("p = \"%s\"\n", p);
		p += strspn (p, " \t");		/* skip white space */
		p += strcspn (p, " \t");	/* skip primary master */
		p += strspn (p, " \t");		/* skip white space */
		p += strcspn (p, " \t");	/* skip mail address */
		dbg_val1 ("p = \"%s\"\n", p);

		dbg_val1 ("is_soa_rr returns = %d\n", (line+len) - p);
		return (line+len) - p;	/* position of serial nr from the end of the line */
	}

	return 0;
}

/*****************************************************************
**	Find string 'search' in 'str' and ignore case in comparison.
**	returns the position of 'search' in 'str' or NULL if not found.
*****************************************************************/
static	const	char	*strfindstr (const char *str, const char *search)
{
	const	char	*p;
	int		c;

	assert ( str != NULL );
	assert ( search != NULL );

	c = tolower (*search);
	p = str;
	do {
		while ( *p && tolower (*p) != c )
			p++;
		if ( strncasecmp (p, search, strlen (search)) == 0 )
			return p;
		p++;
	} while ( *p );

	return NULL;
}

/*****************************************************************
**	return the serial number of the given time in the form
**	of YYYYmmdd00 as unsigned long value
*****************************************************************/
static	unsigned long	serialtime (time_t sec)
{
	struct	tm	*t;
	unsigned long	serialtime;

	t = gmtime (&sec);
	serialtime = (t->tm_year + 1900) * 10000;
	serialtime += (t->tm_mon+1) * 100;
	serialtime += t->tm_mday;
	serialtime *= 100;

	return serialtime;
}

/*****************************************************************
**	inc_soa_serial (fp)
**	increment the soa serial number of the file 'fp'
**	'fp' must be opened "r+"
**	returns 0 on success or a negative value in case of an error
*****************************************************************/
static	int	inc_soa_serial (FILE *fp)
{
	int	c;
	long	pos,	eos;
	unsigned long	serial;
	int	digits;
	unsigned long	today;

	/* move forward until any non ws is reached */
	while ( (c = getc (fp)) != EOF && isspace (c) )
		;
	ungetc (c, fp);		/* push back the last char */

	pos = ftell (fp);	/* mark position */

	serial = 0L;	/* read in the current serial number */
	/* be aware of the trailing space in the format string !! */
	if ( fscanf (fp, "%lu ", &serial) != 1 )	/* try to get serial no */
		return -3;
	eos = ftell (fp);	/* mark first non digit/ws character pos */

	digits = eos - pos;
	if ( digits < 10 )	/* not enough space for serial no ? */
		return -4;

	today = time (NULL);	/* get the current time in seconds since 1/1/1970 */

	switch ( guess_serialfmt (serial) )
	{
	case 'i':	/* plain integer */
		break;
	case 'd':	/* YYYYmmdd00 format ? */
		today = serialtime (today);	/* convert current time in date format */
		if ( serial < today )
			serial = today;
		break;
	case 's':	/* unix time (seconds since "the epoch" ) */
		serial = today;		/* use unixtime */
		break;
	}
	serial++;			/* increment anyway ! */

	fseek (fp, pos, SEEK_SET);	/* go back to the beginning */
	fprintf (fp, "%-*lu", digits, serial);	/* write as many chars as before */

	return 0;	/* yep! */
}

/*****************************************************************
**	return the error text of the inc_serial return coode
*****************************************************************/
static	const	char	*inc_errstr (int err)
{
	switch ( err )
	{
	case -1:	return "couldn't open zone file for modifying";
	case -2:	return "unexpected end of file";
	case -3:	return "no serial number found in zone file";
	case -4:	return "not enough space left for serialno";
	case -5:	return "error on closing zone file";
	}
	return "";
}

/****************************************************************
**
**	int	read_serial_fromfile (filename)
**
**	This function depends on a special syntax formating the
**	SOA record in the zone file!!
**
**	To match the SOA record, the SOA RR must be formatted
**	like this:
**	@    IN  SOA <master.fq.dn.> <hostmaster.fq.dn.> (
**	<SPACEes or TABs>      1234567890; serial number 
**	<SPACEes or TABs>      86400	 ; other values
**				...
**
****************************************************************/
static	int	read_serial_fromfile (const char *fname, unsigned long *serial)
{
	FILE	*fp;
	char	buf[4095+1];
	char	master[254+1];
	int	c;
	int	ret;
	int	soafound;

	if ( (fp = fopen (fname, "r")) == NULL )
		return -1;		/* file not found */

		/* read until the line matches the beginning of a soa record ... */
	ret = 0;
	soafound = 0;
	while ( !soafound && fgets (buf, sizeof buf, fp) )
	{
		if ( sscanf (buf, "%*s %*d IN SOA %255s %*s (\n", master) == 1 )
			soafound = 1;
		else if ( sscanf (buf, "%*s IN SOA %255s %*s (\n", master) == 1 )
			soafound = 1;
	}

	if ( soafound )
	{
		/* move forward until any non ws is reached */
		while ( (c = getc (fp)) != EOF && isspace (c) )
			;
		ungetc (c, fp);		/* pushback the non ws */

		/* read in the current serial number */
		*serial = 0L;
		if ( fscanf (fp, "%lu", serial) != 1 )	/* try to get serial no */
			ret = -3;	/* no serial number found */
	}
	else
		ret = -2;	/* no zone file (soa not found) */

	fclose (fp);

	return ret;
}

/*****************************************************************
**	guess_serialfmt(serial)
*****************************************************************/
static	char	guess_serialfmt (unsigned long serial)
{
	/* try to guess the soa serial format */
	if ( serial < 1136070000L )	/* plain integer (this is 2006-1-1 00:00 in unixtime format) */
		return 'i';
	if ( serial > 2006010100L )	/* date format */
		return 'd';

	return 's';			/* unixtime (seconds) */
}

/*****************************************************************
**	printserial()
*****************************************************************/
static	void	printserial (const char *fname, unsigned long serial)
{
	int	y,	m,	d,	v;

	if ( fname && *fname )
		printf ("%-30s\t", fname);

	printf ("%10lu", serial);

	switch ( guess_serialfmt (serial) )
	{
	case 'i':	/* plain integer */
		break;
		;
	case 'd':	/* date format */
		v = serial % 100;
		serial /= 100;
		d = serial % 100;
		serial /= 100;
		m = serial % 100;
		serial /= 100;
		y = serial;
		
		printf ("\t%d-%02d-%02d Version %02d", y, m, d, v);
		break;
	case 's':	/* unix time (seconds) */
		printf ("\t%s\n", time2str (serial, 's') );
		break;
	}
	printf ("\n");
}

/*****************************************************************
**	usage (msg)
*****************************************************************/
static	void	usage (const char *msg)
{
	if ( msg && *msg )
		fprintf (stderr, "%s\n", msg);
	fprintf (stderr, "usage: %s -s serial\n", progname);
	fprintf (stderr, "usage: %s [-u] zonefile [...]\n", progname);

	exit (1);
}

/*****************************************************************
**	main()
*****************************************************************/
int	main (int argc, char *argv[])
{
	unsigned long	serial;
	int	update;
	int	err;

	progname = *argv;

	if ( --argc == 0 )
		usage ("");

	update = 0;
	if ( argv[1][0] == '-' )
	{
		if ( argv[1][1] == 'u' )
		{
			update = 1;
			argc--;
			argv++;
		}
		else if ( argv[1][1] == 's' )
		{
			if ( argc != 2 )
				usage ("Option -s requires an argument");

			serial = atol (argv[2]);
			printserial ("", serial);
			return 0;
		}
		else
			usage ("illegal option");
	}

	while ( argc-- > 0 )
		if ( update )
		{
			err = inc_serial (*++argv);
			if ( err < 0 )
			{
				fprintf (stderr, "Can not increment serial number in %s\n", *argv);
				fprintf (stderr, "error %d:  %s\n", err, inc_errstr (err));
			}
		}
		else
		{
			if ( (read_serial_fromfile (*++argv, &serial)) != 0 )
				fprintf (stderr, "couldn't read serial number from file %s\n", *argv);
			else
				printserial (*argv, serial);
		}

	return 0;
}
